#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

# Python Modules
import logging
import time

from typing import Optional, Protocol

# 3rd Party Modules
import pytest

from scrapy import Spider
from scrapy.crawler import Crawler
from scrapy.http import HtmlResponse, Request
from scrapy.settings import Settings
from scrapy.utils.python import to_bytes
from w3lib.http import headers_raw_to_dict

# Project Modules
from scrachy.middleware.httpcache import AlchemyCacheStorage
from scrachy.settings.defaults.storage import RetrievalMethod
from test.utils import is_postgresql_setup

log = logging.getLogger("test_storage")


DATABASE_DIALECTS = ['sqlite', 'postgresql'] if is_postgresql_setup() else ['sqlite']


class SettingsFactory(Protocol):
    def __call__(self, hasher: Optional[str], extractor: Optional[str], method: RetrievalMethod) -> Settings:
        pass


@pytest.fixture
def headers(request: pytest.FixtureRequest) -> Optional[str]:
    if request.param: # noqa
        return """
            Host: code.tutsplus.com
            User-Agent: Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US; rv:1.9.1.5) Gecko/20091102 Firefox/3.5.5 (.NET CLR 3.5.30729)
            Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8
            Accept-Language: en-us,en;q=0.5
            Accept-Encoding: gzip,deflate
            Accept-Charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7
            Keep-Alive: 300
            Connection: keep-alive
            Cookie: PHPSESSID=r2t5uvjq435r4q7ib3vtdjq120
            Pragma: no-cache
            Cache-Control: no-cache
        """

    return None


@pytest.fixture
def request_body(request: pytest.FixtureRequest) -> Optional[str]:
    if request.param: # noqa
        return """
            POST /test HTTP/1.1
            Host: foo.example
            Content-Type: application/x-www-form-urlencoded
            Content-Length: 27
        
            field1=value1&field2=value2
        """

    return None


@pytest.fixture
def response_body(request: pytest.FixtureRequest) -> str:
    if request.param == 'page_1':  # noqa
        return "<!-- http://help.websiteos.com/websiteos/example_of_a_simple_html_page.htm -->\n<HTML>\n<HEAD>\n  <TITLE>Your Title Here</TITLE>\n</HEAD>\n<BODY BGCOLOR=\"FFFFFF\">\n<CENTER><IMG SRC=\"clouds.jpg\" ALIGN=\"BOTTOM\"></CENTER>\n<HR>\n<a href=\"http://somegreatsite.com\">Link Name</a>\nis a link to another nifty site\n<H1>This is a Header</H1>\n<H2>This is a Medium Header</H2>\nSend me mail at <a href=\"mailto:support@yourcompany.com\">support@yourcompany.com</a>.\n<P> This is a new paragraph!</P>\n<P><B>This is a new paragraph!</B>\n  <BR> <B><I>This is a new sentence without a paragraph break, in bold italics.</I></B>\n</P>\n<HR>\n</BODY>\n</HTML>"


@pytest.fixture
def extracted_text(request: pytest.FixtureRequest) -> Optional[str]:
    if request.param == ('page_1', 'html.parser'):  # noqa
        return 'Your Title Here\nLink Name\nis a link to another nifty site\nThis is a Header\nThis is a Medium Header\nSend me mail at\nsupport@yourcompany.com\n.\nThis is a new paragraph!\nThis is a new paragraph!\nThis is a new sentence without a paragraph break, in bold italics.'
    elif request.param == ('page_1', 'ArticleExtractor'):  # noqa
        return 'Link Name is a link to another nifty site\nThis is a Header\nThis is a Medium Header\nSend me mail at support@yourcompany.com .\nThis is a new paragraph!\nThis is a new paragraph!\nThis is a new sentence without a paragraph break, in bold italics.'

    return None


@pytest.fixture
def settings_factory(settings_choices: dict[str, Settings]):
    def wrapper(
            hasher: Optional[str],
            extractor: Optional[str],
            retrieval_method: RetrievalMethod = 'full'
    ) -> Settings:
        settings = settings_choices['defaults'].copy()

        if hasher is not None:
            finger_class = 'scrachy.utils.request.DynamicHashRequestFingerprinter'
            settings.set('REQUEST_FINGERPRINTER_CLASS', finger_class)
            settings.set('SCRACHY_REQUEST_FINGERPRINTER_HASHER_CLASS', hasher)

        if extractor is not None:
            if extractor in ('html.parser', 'lxml', 'html5lib'):
                settings.set('SCRACHY_CONTENT_EXTRACTOR', 'scrachy.content.bs4.BeautifulSoupExtractor')
                settings.set('SCRACHY_CONTENT_BS4_PARSER', extractor)
            elif extractor in ('DefaultExtractor', 'ArticleExtractor'):
                settings.set('SCRACHY_CONTENT_EXTRACTOR', 'scrachy.content.boilerpipe.BoilerpipeExtractor')
                settings.set('SCRACHY_BOILERPY_EXTRACTOR', f"boilerpy3.extractors.{extractor}")

        settings.set('SCRACHY_CACHE_RESPONSE_RETRIEVAL_METHOD', retrieval_method)
        settings.set('HTTPCACHE_EXPIRATION_SECS', 0.5)

        return settings

    return wrapper


class MockSpider(Spider):
    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)


@pytest.mark.parametrize(
    'url, hasher, extractor, method, headers, request_body, response_body, extracted_text, wait',
    [
        ('https://www.example.com', None, None, 'minimal', False, False, 'page_1', None, 0.0),
        ('https://www.example.com', None, 'html.parser', 'full', False, False, 'page_1', ('page_1', 'html.parser'), 0.0),
        ('https://www.example.com', None, 'ArticleExtractor', 'full', False, False, 'page_1', ('page_1', 'ArticleExtractor'), 0.0),
        ('https://www.example.com', 'hashlib.md5', 'ArticleExtractor', 'full', True, True, 'page_1', ('page_1', 'ArticleExtractor'), 1.0),
    ],
    indirect=['headers', 'request_body', 'response_body', 'extracted_text']
)
def test_storage(
        settings_factory: SettingsFactory,
        url: str,
        hasher: str,
        extractor: str,
        method: RetrievalMethod,
        headers: Optional[str],
        request_body: Optional[str],
        response_body: str,
        extracted_text: Optional[str],
        wait: float
):
    settings = settings_factory(hasher, extractor, method)
    crawler = Crawler(MockSpider, settings)
    crawler._apply_settings()
    spider = MockSpider.from_crawler(crawler, name='mock_spider')

    storage = AlchemyCacheStorage(settings)
    storage.open_spider(spider)

    request = Request(
        url=url,
        body=request_body,
        headers=headers_raw_to_dict(to_bytes(headers)) if headers else None,
        method='POST' if request_body else 'GET'
    )

    response = HtmlResponse(
        url=url,
        body=response_body,
        encoding='utf-8',
        status=200
    )

    storage.store_response(spider, request, response)

    time.sleep(wait)

    act_response = storage.retrieve_response(spider, request)

    if wait > 0.5:
        assert act_response is None
    else:
        assert act_response.url == request.url == response.url
        assert act_response.text == response.text
        assert act_response.extracted_text == extracted_text

    storage.close_spider(spider)



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

# Python Modules
import datetime
import logging
import os
import pathlib

from typing import Any, Optional, TypeVar

# 3rd Party Modules
from dateutil import parser as date_parser
from dotenv import load_dotenv
from scrapy.settings import Settings
from sqlalchemy import create_engine
from sqlalchemy.exc import OperationalError

# Project Modules
from scrachy.utils.db import construct_url


load_dotenv()

log = logging.getLogger(__name__)


T = TypeVar('T')

TEST_DATABASE_NAME = 'scrachy_test'


def load_postgresql_settings() -> Settings:
    sslrootcert: Optional[str] = os.getenv('TEST_SCRACHY_DB_SSL_ROOT_CERT')
    sslcert: Optional[str] = os.getenv('TEST_SCRACHY_DB_SSL_CERT')
    sslkey: Optional[str] = os.getenv('TEST_SCRACHY_DB_SSL_KEY')
    sslmode: Optional[str] = os.getenv('TEST_SCRACHY_DB_SSL_MODE')

    connect_args = {
        'sslrootcert': sslrootcert,
        'sslcert': sslcert,
        'sslkey': sslkey,
        'sslmode': sslmode
    }
    connect_args = {k: v for k, v in connect_args.items() if v}

    settings = Settings({
        'SCRACHY_DB_DIALECT': 'postgresql',
        'SCRACHY_DB_HOST': os.getenv('TEST_SCRACHY_DB_HOST', 'localhost'),
        'SCRACHY_DB_PORT': int(os.getenv('TEST_SCRACHY_DB_PORT', '5432')),
        'SCRACHY_DB_DATABASE': os.getenv('TEST_SCRACHY_DB_DATABASE', 'scrachy_test'),
        'SCRACHY_DB_USERNAME': os.getenv('TEST_SCRACHY_DB_USERNAME', 'scrachy'),
        'SCRACHY_DB_PASSWORD': os.getenv('TEST_SCRACHY_DB_PASSWORD'),
        'SCRACHY_DB_CONNECT_ARGS': connect_args
    })

    return settings


def is_postgresql_setup() -> bool:
    """
    Check to see if it is possible to connect to the ``scrachy_test``
    database using the user ``scrachy``.

    :return:
    """
    try:
        settings = load_postgresql_settings()
    except Exception as e:
        log.error(f"There was a problem configuring the test database settings: {str(e)}")
        return False

    url = construct_url(settings)
    engine = create_engine(url, connect_args=settings['SCRACHY_DB_CONNECT_ARGS'])

    try:
        with engine.connect():
            can_connect = True
    except OperationalError as e:
        log.warning(f"Unable to connect to the postgresql test database: {str(e)}")
        can_connect = False
    finally:
        if engine is not None:
            engine.dispose()

    return can_connect


def update_database_settings(original_settings: Settings | list[Settings], dialect: str, basedir: pathlib.Path):
    if isinstance(original_settings, Settings):
        original_settings = [original_settings]

    for current_settings in original_settings:
        current_settings['SCRACHY_DB_DIALECT'] = dialect

        if dialect in ('sqlite', 'other'):
            db_dir = basedir
            db_dir.mkdir(parents=True, exist_ok=True)
            db_path = db_dir / f'{TEST_DATABASE_NAME}.sqlite3'
            db_path.unlink(missing_ok=True)

            current_settings.set('SCRACHY_DB_DIALECT', 'sqlite')
            current_settings.set('SCRACHY_DB_HOST', None)
            current_settings.set('SCRACHY_DB_PORT', None)
            current_settings.set('SCRACHY_DB_USERNAME', None)
            current_settings.set('SCRACHY_DB_PASSWORD', None)
            current_settings.set('SCRACHY_DB_DATABASE', str(db_path))
            current_settings.set('SCRACHY_DB_SCHEMA', None)
            current_settings.set('SCRACHY_DB_CONNECT_ARGS', {})
        elif dialect == 'postgresql':
            current_settings.update(load_postgresql_settings())


def parse_date(obj: Any) -> datetime.datetime:
    if isinstance(obj, str):
        return date_parser.parse(obj)

    if isinstance(obj, datetime.datetime):
        return obj

    raise ValueError(f"Can't parse: {str(obj)}")



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

# Python Modules
import datetime
import logging
import re

from contextlib import nullcontext as does_not_raise

# 3rd Party Modules
import pytest

from scrapy.crawler import Crawler
from scrapy.exceptions import IgnoreRequest
from scrapy.http import Request
from scrapy.settings import Settings, iter_default_settings

# Project Modules
from scrachy.db.engine import session_scope
from scrachy.db.models import Response
from scrachy.db.repositories import ResponseRepository
from scrachy.middleware.filter import CachedResponseFilter
from scrachy.settings.defaults.storage import PatternLike
from test.integration.middleware.conftest import MockSpider


log = logging.getLogger('test_cache_filter')


CACHED_URLS = [
    'https://www.google.com/page1',
    'https://www.google.com/page2',
    'https://university.edu',
    'https://university.edu/page1',
    'https://example.org'
]

UN_CACHED_URLS = [
    'https://www.google.com/page5',
    'https://college.edu',
    'https://college.edu/page1',
    'https://example.org/page_1'
]


@pytest.fixture
def settings(request: pytest.FixtureRequest) -> Settings:
    exclusions = request.getfixturevalue('exclusions')

    log.debug(f"Exclusions: {exclusions}")

    settings = Settings(dict(iter_default_settings()))
    settings.set('SCRACHY_DB_DIALECT', 'sqlite')
    settings.set('SCRACHY_CACHED_RESPONSE_FILTER_EXCLUSIONS', exclusions)

    return settings


@pytest.fixture
def crawler(request: pytest.FixtureRequest) -> Crawler:
    settings: Settings = request.getfixturevalue('settings')

    crawler = Crawler(
        MockSpider,
        settings=settings
    )

    crawler._apply_settings() # noqa

    return crawler


@pytest.fixture
def populate_cache(request: pytest.FixtureRequest):
    crawler: Crawler = request.getfixturevalue('crawler')
    log.debug(f"Crawler: {crawler}")

    urls = CACHED_URLS

    fingerprints = [crawler.request_fingerprinter.fingerprint(Request(url)) for url in urls]
    scrape_timestamps = [datetime.datetime(2023, 1, 1, 12, i) for i in range(len(urls))]

    with session_scope() as session:
        repo = ResponseRepository(session)

        for fingerprint, scrape_timestamp, url in zip(fingerprints, scrape_timestamps, urls):
            repo.insert(
                Response(
                    fingerprint=fingerprint,
                    scrape_timestamp=scrape_timestamp,
                    url=url,
                    request_method='GET',
                    body='',
                    body_length=0
                ))


@pytest.mark.parametrize('dialect', ['sqlite'])
@pytest.mark.parametrize(
    'exclusions, raise_expectations',
    [
        (
            [r'google\.com', r'\.edu$'],
            [does_not_raise(), does_not_raise(), does_not_raise(), pytest.raises(IgnoreRequest), 
             pytest.raises(IgnoreRequest), does_not_raise(), does_not_raise(), does_not_raise(), does_not_raise()]
        ),
        (
            [re.compile(r'\.edu$')],
            [pytest.raises(IgnoreRequest), pytest.raises(IgnoreRequest), does_not_raise(),
             pytest.raises(IgnoreRequest), pytest.raises(IgnoreRequest), does_not_raise(), does_not_raise(),
             does_not_raise(), does_not_raise()]
        ),
    ]
)
def test_cache_filter(
        settings: Settings,
        crawler: Crawler,
        manage_engine,
        populate_cache,
        dialect: str,
        exclusions: list[PatternLike],
        raise_expectations: list
):
    cache_filter = CachedResponseFilter(crawler)
    spider = crawler.spider

    # It's not (easily) possible to test if an item should be skipped due to
    # being stale. That path of code is fairly straight forward and the
    # expiration manager is tested, so it should be correct.
    for url, raise_expectation in zip(CACHED_URLS + UN_CACHED_URLS, raise_expectations):
        with raise_expectation:
            cache_filter.process_request(Request(url=url), spider)

    dont_cache = {'dont_cache': True}
    dont_filter = {'dont_filter': True}
    cachable_url = 'https://example.org'

    assert cache_filter.process_request(Request(url=cachable_url, meta=dont_cache), crawler.spider) is None
    assert cache_filter.process_request(Request(url=cachable_url, meta=dont_filter), crawler.spider) is None



#  Copyright 2020 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
The basic data types and classes required to define the SqlAlchemy models.
"""

from __future__ import annotations

# Python Modules
import json
import logging

from json import JSONDecodeError
from typing import Annotated, Any, Optional, Sequence

# 3rd Party Modules
from sqlalchemy import MetaData, inspect, BigInteger, LargeBinary, SmallInteger
from sqlalchemy.orm import DeclarativeBase, NO_VALUE, QueryableAttribute, declared_attr

# Project Modules
from scrachy.settings import PROJECT_SETTINGS
from scrachy.utils.sqltypes import TimeStampTZ
from scrachy.utils.strings import camel_to_snake

bigint = Annotated[BigInteger, 64]
binary = Annotated[LargeBinary, None]
smallint = Annotated[int, 16]
timestamp = Annotated[TimeStampTZ, None]

log = logging.getLogger(__name__)


schema = PROJECT_SETTINGS.get('SCRACHY_DB_SCHEMA')
schema_prefix = f"{schema}." if schema else ""


class Base(DeclarativeBase):
    metadata = MetaData(
        naming_convention={
            "ix": "ix_%(column_0_label)s",
            "uq": "uq_%(table_name)s_%(column_0_name)s",
            "ck": "ck_%(table_name)s_%(constraint_name)s",
            "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
            "pk": "pk_%(table_name)s",
        },
        schema=schema
    )

    type_annotation_map = {
        bigint: BigInteger,
        binary: LargeBinary,
        smallint: SmallInteger,
        timestamp: TimeStampTZ(timezone=True),
    }

    # noinspection PyMethodParameters
    @declared_attr
    def __tablename__(cls):
        return camel_to_snake(cls.__name__)

    # Modified from https://stackoverflow.com/a/55749579/4971706
    def __repr__(self) -> str:
        return json.dumps(self.to_dict(), indent=2, sort_keys=True, default=str)

    def __eq__(self, other: Base) -> bool:
        """
        A model is equal to another if all of its columns are equal.

        :param other:
        :return:
        """
        columns: Sequence[str] = self.__table__.columns.keys()

        for col in columns:
            this_value = getattr(self, col)

            try:
                that_value = getattr(other, col)
            except AttributeError:
                return False

            if this_value != that_value:
                return False

        return True

    def __hash__(self):
        d = {k: getattr(self, k) for k in sorted(self.__table__.columns.keys())}

        return hash(tuple(d.items()))

    def to_dict(self, hide_missing: bool = True, exclude_keys: Optional[Sequence[str]] = None):
        # Adapted from: https://medium.com/@alanhamlett/part-1-sqlalchemy-models-to-json-de398bc2ef47
        # The only major change was to add a check to make sure the relationship
        # data is attached before trying to access it with getattr.
        path: str = self.__tablename__.lower()
        excluded: set[str] = set(exclude_keys) if exclude_keys else set()
        seen: set[Any] = {self}

        return self._to_dict(path, excluded, hide_missing, seen)

    def _to_dict(self, path: str, excluded: set[str], hide_missing: bool, seen: set[Any]) -> dict[str, Any]:
        columns: Sequence[str] = self.__table__.columns.keys()
        relationships: Sequence[str] = self.__mapper__.relationships.keys()
        properties: Sequence[str] = list(set(dir(self)) - set(columns) - set(relationships))

        result: dict[str, Any] = dict()

        # The columns
        self._columns_to_dict(columns, path, excluded, result, hide_missing)
        self._relationships_to_dict(relationships, path, excluded, result, hide_missing, seen)
        self._properties_to_dict(properties, path, excluded, result, hide_missing)

        return result

    def _columns_to_dict(
            self,
            columns: Sequence[str],
            path: str,
            excluded: set[str],
            result: dict[str, Any],
            hide_missing: bool
    ):
        for key in columns:
            if self._is_private(key):
                continue

            # Some keys might be an SqlAlchemy subclass of str
            key = str(key)

            qualified_key = self._qualified_key(path, key)
            if qualified_key in excluded:
                continue

            value = getattr(self, key)

            if value is None and hide_missing:
                continue

            result[key] = value.hex() if isinstance(value, bytes) else value

    def _relationships_to_dict(
            self,
            relationships: Sequence[str],
            path: str,
            excluded: set[str],
            result: dict[str, Any],
            hide_missing: bool,
            seen: set[Any]
    ):
        for key in relationships:
            if self._is_private(key):
                continue

            qualified_key = self._qualified_key(path, key)
            if qualified_key in excluded:
                continue

            key = str(key)

            excluded.add(qualified_key)

            relationship = self.__mapper__.relationships[key]

            if relationship.uselist:
                items: list[Base] = getattr(self, key)
                if relationship.query_class is not None:
                    if hasattr(items, "all"):
                        items = items.all()
                result[key] = [i._to_dict(qualified_key.lower(), excluded, hide_missing, seen) for i in items]
            else:
                if relationship.query_class is not None or relationship.instrument_class is not None:
                    state = inspect(self)
                    loaded_value = state.attrs[key].loaded_value
                    if loaded_value == NO_VALUE:
                        result[key] = '[DETACHED]'
                    else:
                        item: Base = getattr(self, key)

                        if item is not None and item not in seen:
                            seen.add(item)
                            result[key] = item._to_dict(qualified_key.lower(), excluded, hide_missing, seen)
                        elif not hide_missing:
                            result[key] = None
                else:
                    value = getattr(self, key)
                    if (value is not None or not hide_missing) and value not in seen:
                        seen.add(value)
                        result[key] = value

    def _properties_to_dict(
            self,
            properties: Sequence[str],
            path: str,
            excluded: set[str],
            result: dict[str, Any],
            hide_missing: bool
    ):
        for key in properties:
            if self._is_private(key):
                continue

            mycls = self.__class__
            if not hasattr(mycls, key):
                continue

            attr = getattr(mycls, key)
            if not isinstance(attr, property) or isinstance(attr, QueryableAttribute):
                continue

            qualified_key = self._qualified_key(path, key)
            if qualified_key in excluded:
                continue

            key = str(key)

            value = getattr(self, key)
            if hasattr(value, '_to_dict'):
                result[key] = value._to_dict(qualified_key.lower(), excluded, hide_missing)
            else:
                try:
                    result[key] = json.loads(json.dumps(value, sort_keys=True))
                except (RecursionError, ValueError, TypeError, JSONDecodeError):
                    pass

    @staticmethod
    def _qualified_key(path: str, key: str) -> str:
        return f"{path}.{key}"

    @staticmethod
    def _is_private(key: str) -> bool:
        return key.startswith('_')



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

# Python Modules
import copy
import logging

from collections import defaultdict

# 3rd Party Modules
import pytest

from dotenv import load_dotenv
from scrapy.settings import Settings

# Project Modules
from scrachy.db.engine import session_scope
from scrachy.db.models import Response as CachedResponse, ScrapeHistory
from scrachy.db.repositories import ResponseRepository, ScrapeHistoryRepository


load_dotenv()


log = logging.getLogger('test_repositories')

log.debug(f"From test_repositories")


@pytest.fixture
def settings(settings_choices: dict[str, Settings], request: pytest.FixtureRequest) -> Settings:
    return settings_choices[request.param]  # noqa


# Note: It's difficult (maybe impossible?) to test if insert/upsert is
# working without also testing find and vice versa.
@pytest.mark.parametrize('dialect', ['sqlite', 'postgresql', 'other'])
@pytest.mark.parametrize('settings', ['defaults', 'settings_2'], indirect=True)
def test_upsert_and_find(
        manage_engine,
        dialect: str,
        settings: Settings,
        cached_responses: list[CachedResponse]
):
    # Create a copy because the objects could be modified by the upsert
    # process.
    cached_responses_copy = copy.deepcopy(cached_responses)

    with session_scope() as session:
        response_repo = ResponseRepository(session)
        history_repo = ScrapeHistoryRepository(session)

        response_repo.dialect = dialect

        for response in cached_responses_copy:
            history = ScrapeHistory(
                fingerprint=response.fingerprint,
                scrape_timestamp=response.scrape_timestamp,
                body=response.body
            )

            act_response = response_repo.upsert(response, returning=True)
            history_repo.insert(history)

            assert act_response == response

    with session_scope() as session:
        response_repo = ResponseRepository(session)
        response_repo.dialect = dialect

        sorted_responses = sorted(
            cached_responses,
            key=lambda e: (e.fingerprint, e.scrape_timestamp),
            reverse=True
        )
        expected_responses = []
        fingers = defaultdict(list)
        for r in sorted_responses:
            if r.fingerprint not in fingers:
                expected_responses.append(r)

            fingers[r.fingerprint].append(r.body)

        # One at a time
        assert all([e == response_repo.find_by_fingerprint(e.fingerprint) for e in expected_responses])

        # All of them
        act_responses = response_repo.find_all()
        assert len(expected_responses) == len(act_responses)
        assert all([e == a for e, a in zip(expected_responses, act_responses)])

        # Make sure the scrape history has the correct number of entries
        for act_response in act_responses:
            act_history = sorted(act_response.scrape_history, key=lambda h: h.scrape_timestamp, reverse=True)
            exp_history = fingers[act_response.fingerprint]

            assert len(exp_history) == len(act_history)
            assert all([exp == act.body for exp, act in zip(exp_history, act_history)])



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

# Python Modules
import logging
import pathlib

# 3rd Party Modules
import pytest

from scrapy.settings import Settings


# Project Modules
from scrachy.db.base import Base
from scrachy.db.engine import initialize_engine, reset_engine
from test.utils import update_database_settings

log = logging.getLogger(__name__)


@pytest.fixture
def manage_engine(request: pytest.FixtureRequest, tmp_path: pathlib.Path):
    dialect: str = request.getfixturevalue('dialect')
    settings: Settings = request.getfixturevalue('settings')

    update_database_settings(settings, dialect, tmp_path)

    engine = initialize_engine(settings)

    yield

    Base.metadata.drop_all(engine)
    reset_engine()



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

# Python Modules
import logging
from contextlib import AbstractContextManager, nullcontext as does_not_raise
from typing import Optional

# 3rd Party Modules
import pytest
import pytest_twisted

from scrapy.crawler import Crawler
from scrapy.http import Request
from scrapy.settings import Settings, iter_default_settings

# Project Modules
from scrachy.http_ import SeleniumRequest
from scrachy.middleware.selenium import AsyncSeleniumMiddleware
from scrachy.settings.defaults import selenium as selenium_defaults
from scrachy.settings.defaults.selenium import WebDriverName
from test.integration.middleware.conftest import MockSpider, SettingsFactory
from test.integration.middleware.test_selenium import paginate

log = logging.getLogger("test_async_selenium")


@pytest.fixture(scope='module')
def middleware() -> AsyncSeleniumMiddleware:
    settings = Settings(dict(iter_default_settings()))
    settings.setmodule(selenium_defaults)
    settings.set('CONCURRENT_REQUESTS', 4)  # temporary
    settings.set('SCRACHY_SELENIUM_WEB_DRIVER', 'Chrome')
    settings.set('SCRACHY_SELENIUM_WEB_DRIVER_OPTIONS', ['--headless=new'])

    crawler = Crawler(
        spidercls=MockSpider,
        settings=settings
    )
    middleware = AsyncSeleniumMiddleware.from_crawler(crawler)

    yield middleware

    log.info(f"Calling spider_closed")
    middleware.spider_closed(crawler.spider)


@pytest_twisted.ensureDeferred
@pytest.mark.parametrize(
    'driver, options, extensions, raises_expectation',
    [
        (None, None, None, does_not_raise()),
        ('Chrome', ['--headless=new'], True, does_not_raise()),
        ('Firefox', ['-headless'], True, does_not_raise()),
    ],
    indirect=['extensions']
)
async def test_initialize_driver(
        settings_factory: SettingsFactory,
        driver: WebDriverName,
        options: Optional[list[str]],
        extensions: Optional[list[str]],
        raises_expectation: AbstractContextManager
):
    with raises_expectation:
        settings = settings_factory(driver, options, extensions)

        crawler = Crawler(
            spidercls=MockSpider,
            settings=settings
        )

        try:
            middleware = AsyncSeleniumMiddleware.from_crawler(crawler)
            assert len(middleware.drivers) == settings.getint('CONCURRENT_REQUESTS')
        except:  # noqa
            pass
        finally:
            middleware.spider_closed(crawler.spider)


@pytest_twisted.ensureDeferred
async def test_return_none_on_scrapy_request(middleware: AsyncSeleniumMiddleware):
    scrapy_request = Request(url='http://not-an-url')

    assert middleware.process_request(scrapy_request) is None


@pytest_twisted.ensureDeferred
async def test_return_response(middleware: AsyncSeleniumMiddleware):
    request = SeleniumRequest(url='https://www.scrapethissite.com/pages/forms/')
    response = await middleware.process_request(request)

    title = response.css('title::text').get()
    assert 'Hockey Teams' in title


@pytest_twisted.ensureDeferred
async def test_screenshot(middleware: AsyncSeleniumMiddleware):
    request = SeleniumRequest(
        url='https://www.scrapethissite.com/pages/forms/',
        screenshot=True
    )
    response = await middleware.process_request(request)

    assert response.meta['screenshot'] is not None


@pytest_twisted.ensureDeferred
async def test_script(middleware: AsyncSeleniumMiddleware):
    request = SeleniumRequest(
        url='https://www.scrapethissite.com/pages/ajax-javascript/',
        script_executor=paginate
    )

    response = await middleware.process_request(request)
    expected_titles = {
        '2015': 'Spotlight',
        '2014': 'Birdman',
        '2013': '12 Years a Slave',
        '2012': 'Argo',
        '2011': 'The Artist',
        '2010': 'The King\'s Speech'
    }

    for key, script_response in response.meta['script_result'].items():
        title = script_response.css('tbody#table-body > tr.film > td.film-title::text').get().strip()

        assert expected_titles[key] == title



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

# Python Modules
import datetime
import logging

# 3rd Party Modules
import pytest

from scrapy.settings import Settings

# Project Modules
from scrachy.middleware.httpcache import ExpirationManager
from test.utils import parse_date


log = logging.getLogger('test_expiration_manager')


@pytest.fixture
def settings(settings_choices: dict[str, Settings], request: pytest.FixtureRequest) -> Settings:
    return settings_choices[request.param]  # noqa


@pytest.fixture
def scrape_timestamp(request: pytest.FixtureRequest) -> datetime.datetime:
    return parse_date(request.param).replace(tzinfo=datetime.timezone.utc) # noqa


@pytest.fixture
def current_timestamp(request: pytest.FixtureRequest) -> datetime.datetime:
    return parse_date(request.param).replace(tzinfo=datetime.timezone.utc) # noqa


@pytest.mark.parametrize(
    'settings, url, scrape_timestamp, current_timestamp, expected_staleness',
    [
        ('defaults', 'http://www.example.com', '2020-01-01T12:00', '2025-01-01T12:00', False),
        ('defaults', 'ftp://anything.edu', '2020-01-01T12:00', '2025-01-01T12:00', False),

        ('defaults_with_expiration_secs', 'http://www.example.com', '2020-01-01T12:01', '2020-01-01T12:02', False),
        ('defaults_with_expiration_secs', 'http://www.example.com', '2020-01-01T12:01', '2020-01-01T12:03', True),
        ('defaults_with_activation_secs', 'http://www.example.com', '2020-01-01T12:01', '2020-01-01T12:02', True),
        ('defaults_with_activation_secs', 'http://www.example.com', '2020-01-01T12:01', '2020-01-01T12:03', False),

        ('defaults_with_expiration_schedule', 'http://www.example.com', '2020-01-01T12:01', '2020-01-01T13:00', False),
        ('defaults_with_expiration_schedule', 'http://www.example.com', '2020-01-01T12:01', '2020-01-02T13:01', True),

        ('defaults_with_activation_pat', 'http://www.example.com', '2020-01-01T12:01', '2020-01-01T12:02', True),
        ('defaults_with_activation_pat', 'https://www.example.com', '2020-01-01T12:01', '2020-01-01T12:02', True),
        ('defaults_with_activation_pat', 'https://www.example.com', '2020-01-01T12:01', '2020-01-01T12:03', False),
        ('defaults_with_activation_pat', 'ftp://www.example.com', '2020-01-01T12:01', '2020-01-01T12:02', False),
        
        ('defaults_with_expiration_pat', 'http://www.example.com', '2020-01-01T12:01', '2020-01-01T12:02', False),
        ('defaults_with_expiration_pat', 'https://www.example.com', '2020-01-01T12:01', '2020-01-01T12:02', False),
        ('defaults_with_expiration_pat', 'https://www.example.com', '2020-01-01T12:01', '2020-01-01T12:03', True),
        ('defaults_with_expiration_pat', 'ftp://www.example.com', '2020-01-01T12:01', '2020-01-01T12:02', False),

        ('defaults_with_expiration_schedule_pat', 'http://www.example.com', '2020-01-01T12:01', '2020-01-01T13:01', False),
        ('defaults_with_expiration_schedule_pat', 'https://www.example.com', '2020-01-01T12:01', '2020-01-02T12:01', True),
        ('defaults_with_expiration_schedule_pat', 'ftp://www.example.com', '2020-01-01T12:01', '2020-01-02T12:01', False),

        ('all_expiration_values', 'http://test.com', '2020-01-01T12:01', '2020-01-02T12:01:05', True),  # Non matching
        ('all_expiration_values', 'http://test.com', '2020-01-01T12:01', '2020-01-01T12:01:20', False), # Non matching
        ('all_expiration_values', 'http://test.com', '2020-01-01T12:01', '2020-01-01T12:18:00', True),  # Non matching

        # This has been in the cache long enough to be active, but not long enough to exceed expiration_secs,
        # but it should expire due to the default schedule.
        ('all_expiration_values', 'http://test.com', '2020-01-29T23:30', '2020-01-31T00:00:00', True),  # Non matching

        ('all_expiration_values', 'http://www.example.com', '2020-01-01T12:01', '2020-01-01T12:00:40', True),
        ('all_expiration_values', 'http://www.example.com', '2020-01-01T12:01', '2020-01-01T12:03:00', False),
        ('all_expiration_values', 'http://www.example.com', '2020-01-01T12:01', '2020-01-01T12:10:00', True),

        ('all_expiration_values', 'http://www.example.com', '2020-01-01T11:55', '2020-01-01T12:00:01', True),

        ('all_expiration_values', 'https://www.example.com', '2020-01-01T12:01', '2020-01-01T12:02:40', True),
        ('all_expiration_values', 'https://www.example.com', '2020-01-01T12:01', '2020-01-01T13:10:00', False),
        ('all_expiration_values', 'https://www.example.com', '2020-01-01T12:01', '2020-01-01T14:00:00', True),
        ('all_expiration_values', 'https://www.example.com', '2020-01-01T11:50', '2020-01-01T12:00:01', False),

        ('all_expiration_values', 'ftp://www.example.com', '2020-01-01T11:59', '2020-01-01T12:00:01', True),
    ],
    indirect=['settings', 'scrape_timestamp', 'current_timestamp']
)
def test_manager(
        settings: Settings,
        url: str,
        scrape_timestamp: datetime.datetime,
        current_timestamp: datetime.datetime,
        expected_staleness: bool
):
    manager = ExpirationManager(settings)

    assert manager.is_stale(url, scrape_timestamp, current_timestamp) == expected_staleness



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
Utilities for managing dynamic imports.
"""

# Python Modules
import inspect

from typing import Any

# 3rd Party Modules

# Project Modules


def get_import_path(obj: Any) -> str:
    """
    Get the import path to the given object. If the ``obj`` is a string then
    we assume it is already an import path and return the value unchanged.

    :param obj:
    :return:
    """

    if isinstance(obj, str):
        return obj

    obj_module = inspect.getmodule(obj)

    return f"{obj_module.__name__}.{obj.__name__}"



#  Copyright 2020 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
The SqlAlchemy models.
"""

from __future__ import annotations

# Python Modules
from typing import Annotated, Optional

# 3rd Party Modules
import sqlalchemy
import sqlalchemy.orm

from sqlalchemy import ColumnElement, String, ForeignKey, type_coerce
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import Mapped, mapped_column, relationship

# Project Modules
from scrachy.db.base import Base, binary, smallint, timestamp
from scrachy.settings import PROJECT_SETTINGS

# TODO I'm not sure the schem and qualified table name stuff is required,
#      because of the 'schema_translate_map` added to the engine
#      execution_options.
schema = PROJECT_SETTINGS.get('SCRACHY_DB_SCHEMA')
schema_prefix = f"{schema}." if schema else ""

str_pk = Annotated[str, mapped_column(primary_key=True)]
int_pk = Annotated[int, mapped_column(primary_key=True)]


def qtn(name: str) -> str:
    """
    Return the qualified table name (i.e., with schema if necessary).

    :param name:
    :return:
    """
    return f"{schema_prefix}{name}"


# Primary tables
class Response(Base):
    """
    Stores the minimum amount of information to recreate the html of a
    page given its request fingerprint. The cache only stores the most
    recent version of the page, but the :class:`ScrapyHistory` table
    can be used to view

    Additional data is stored in other tables, but can be accessed via
    relationships.
    """
    # A hash of the page's uniquely identifying request information.
    fingerprint: Mapped[binary] = mapped_column(primary_key=True)

    # The most recent timestamp when the page was scraped.
    scrape_timestamp: Mapped[timestamp]

    # This is redundant with the request, but can prevent a join when
    # we might want to search or filter responses by it.
    url: Mapped[str] = mapped_column(index=True)

    # The request method (True for GET. False for POST)
    method: Mapped[bool]

    # The body of the request (primarily used for POST requests)
    request_body: Mapped[Optional[str]]

    # The response body
    body: Mapped[str]

    # The response status code for the page
    status: Mapped[Optional[smallint]]

    # The response headers
    headers: Mapped[Optional[str]]

    # The text of the response stripped of HTML and possibly stripped of
    # boilerplate content.
    extracted_text: Mapped[Optional[str]]

    # The number of bytes of the full downloaded response (including the HTML)
    body_length: Mapped[int]

    # The number of bytes of the response excluding the HTML.
    extracted_text_length: Mapped[Optional[int]]

    # The scrape history
    scrape_history: Mapped[list[ScrapeHistory]] = relationship(cascade='all, delete-orphan')

    # I believe hybrid properties allow changing how the property behaves
    # depending on whether you are using it as an instance or as a class
    # template. For example, when operating on an instance referencing the
    # `body` property will use the regular python code in the method body.
    # But if you using it as a template for SQL operations you can define
    # a @xxx.inplace.expression method that will use the same property name
    # but emit SQL code instead.

    # So here, for example, it uses less space to store the request method
    # as a boolean (because it can only ever have two values), but it is
    # more convenient to work with these values as their original string
    # representations.
    @hybrid_property
    def request_method(self) -> Optional[str]:
        if self.method is True:
            return 'GET'
        elif self.method is False:
            return 'POST'

        return None

    @request_method.inplace.setter
    def _request__method(self, method: str) -> None:
        self.method = method.upper() == 'GET'  # noqa

    # noinspection PyMethodParameters,PyNestedDecorators
    @request_method.inplace.expression
    @classmethod
    def _request_method(cls) -> ColumnElement[Optional[str]]:
        return type_coerce(
            sqlalchemy.case(
                (cls.method == True, 'GET'),
                (cls.method == False, 'POST'),
                else_=None
            ),
            String
        )


class ScrapeHistory(Base):
    """
    Store the history of when a page was scraped.
    """
    fingerprint: Mapped[binary] = mapped_column(ForeignKey(qtn('response.fingerprint')), primary_key=True)

    # When the page was scraped
    scrape_timestamp: Mapped[timestamp] = mapped_column(primary_key=True)

    # The response body
    body: Mapped[str]



#  Copyright 2020 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.
"""
Additional ``Request`` and ``Response`` classes for working with Selenium
and the ``AlchemyCacheStorage`` backend.

Note: Naming this module ``http`` causes a circular import error, so I've appended
an underscore to avoid conflicts.
"""
from __future__ import annotations

# Python Modules
import datetime
import logging

from typing import Any, Optional, Protocol

# 3rd Party Modules
from scrapy.http import HtmlResponse, TextResponse, Request, Response, XmlResponse
from selenium.webdriver.remote.webdriver import WebDriver

# Project Modules
from scrachy.db.models import ScrapeHistory

log = logging.getLogger(__name__)


class WaitCondition(Protocol):
    def __call__(self, driver: WebDriver) -> Any:
        pass


class ScriptExecutor(Protocol):
    def __call__(self, driver: WebDriver, request: Request) -> Optional[Response | list[Response] | dict[str, Response]]:
        pass


class SeleniumRequest(Request):
    """
    A subclas of :class:`scrapy.http.Request` that provides extra information for downloading pages using
    Selenium.

    Based off the code from `Scrapy-Selenium <https://github.com/clemfromspace/scrapy-selenium>`_
    """

    def __init__(
            self,
            wait_timeout: Optional[float] = None,
            wait_until: Optional[WaitCondition] = None,
            screenshot: bool = False,
            script_executor: Optional[ScriptExecutor] = None,
            *args,
            **kwargs
    ):
        """
        A new ``SeleniumRequest``.

        :param wait_timeout: The number of seconds to wait before accessing the data.
        :param wait_until: One of the "selenium.webdriver.support.expected_conditions". The response
                           will be returned until the given condition is fulfilled.
        :param screenshot: If ``True``, a screenshot of the page will be taken and the data of the screenshot
                           will be returned in the response "meta" attribute.
        :param script_executor: A function that takes a webdriver and a response as its parameters and optionally
                                returns a list of new response objects as a side effect of its actions (e.g.,
                                executing arbitrary javascript code on the page). Any returned responses will
                                be returned in the ``request.meta`` attribute with the key ``script_result``.
                                Note that the returned responses will not be further processed by any other
                                middleware.

        """
        super().__init__(*args, **kwargs)

        self.wait_timeout = wait_timeout
        self.wait_until = wait_until
        self.screenshot = screenshot
        self.script_executor = script_executor


class CachedResponseMixin:
    def __init__(
            self,
            scrape_timestamp: Optional[datetime.datetime] = None,
            extracted_text: Optional[str] = None,
            body_length: Optional[int] = None,
            extracted_text_length: Optional[int] = None,
            scrape_history: Optional[list[ScrapeHistory]] = None,
            *args,
            **kwargs
    ):
        """
        A subclass of :class:`scrapy.http.HttpResponse` that contains a
        subset of the extra information stored in the cache.

        :param scrape_timestamp: The most recent date the request was scraped.
        :param body_number_of_bytes: The total number of bytes of the downloaded
               html.
        :param text_number_of_bytes: The number of bytes in the extracted
               plain text.
        :param body_text: The text extracted from the HTML.
        """
        super().__init__(*args, **kwargs)

        self.scrape_timestamp = scrape_timestamp
        self.extracted_text = extracted_text
        self.body_length = body_length
        self.extracted_text_length = extracted_text_length
        self.scrape_history = scrape_history


class CachedHtmlResponse(CachedResponseMixin, HtmlResponse):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class CachedTextResponse(CachedResponseMixin, TextResponse):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class CachedXmlResponse(CachedResponseMixin, XmlResponse):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

# Python Modules
import pathlib

from typing import Any, Optional, Protocol

# 3rd Party Modules
import pytest

from scrapy import Spider
from scrapy.http import Response
from scrapy.settings import Settings, iter_default_settings

# Project Modules
from scrachy.settings.defaults import filter as filter_defaults
from scrachy.settings.defaults import selenium as selenium_defaults
from scrachy.settings.defaults.selenium import WebDriverName
from scrachy.settings.defaults.storage import PatternLike


class MockSpider(Spider):
    name = 'mock_spider'
    allowed_domains = ['scrapethissite.com']

    def parse(self, response: Response, **kwargs: Any) -> Any:
        pass


class SettingsFactory(Protocol):
    def __call__(
            self,
            driver_name: Optional[WebDriverName] = None,
            options: Optional[list[str]] = None,
            extensions: Optional[list[str]] = None
    ) -> Settings:
        pass


@pytest.fixture
def extensions(request: pytest.FixtureRequest) -> list[str]:
    if not request.param:  # noqa
        return []

    driver_name = request.node.funcargs['driver']
    basedir = pathlib.Path(__file__).parent / 'selenium' / 'extensions'

    if driver_name == 'Chrome':
        return [str(f) for f in basedir.glob('*.crx')]
    elif driver_name == 'Firefox':
        return [str(f) for f in basedir.glob('*.xpi')]

    raise ValueError(f"Unsupported driver: {driver_name}")


@pytest.fixture
def settings_factory() -> SettingsFactory:
    settings = Settings(dict(iter_default_settings()))
    settings.setmodule(selenium_defaults)
    settings.setmodule(filter_defaults)

    def wrapper(
            driver_name: Optional[WebDriverName] = None,
            options: Optional[list[str]] = None,
            extensions: Optional[list[str]] = None,
            exclusions: Optional[list[PatternLike]] = None,
    ) -> Settings:
        s = settings.copy()
        s['CONCURRENT_REQUESTS'] = 1

        if driver_name:
            s['SCRACHY_SELENIUM_WEB_DRIVER'] = driver_name

        if options:
            s['SCRACHY_SELENIUM_WEB_DRIVER_OPTIONS'] = options

        if extensions:
            s['SCRACHY_SELENIUM_WEB_DRIVER_EXTENSIONS'] = extensions

        s['SCRACHY_CACHED_RESPONSE_FILTER_EXCLUSIONS'] = exclusions or []

        return s

    return wrapper



#  Copyright 2020 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

# Python Modules
import pathlib

from setuptools import find_packages, setup

# 3rd Party Modules

# Project Modules


current_dir = pathlib.Path(__file__).parent
readme = (current_dir / "README.md").read_text()

extras = {
    'content_extraction': ['beautifulsoup4', 'boilerpy3'],
    'html_parsing': ['html5lib', 'lxml'],
    'mysql': ['pymysql'],
    'postgresql': ['psycopg2'],
    'testing': ['pytest', 'pytest-twisted', 'python-dotenv', 'pyyaml']
}
all_extras = [package_name for extra_list in extras.values() for package_name in extra_list]


if __name__ == '__main__':
    setup(
        name='scrachy',
        version='0.5.7',
        description='Enhanced caching modules for scrapy.',
        long_description=readme,
        long_description_content_type='text/markdown',
        install_requires=[
            'cron-converter',
            'msgspec',
            'scrapy',
            'selenium',
            'sqlalchemy',
            'twisted',
            'w3lib',
            'pywin32; os_name=="nt"'
        ],
        extras_require={'all': all_extras} | extras,
        author='Reid Swanson',
        maintainer='Reid Swanson',
        author_email='reid@reidswanson.com',
        maintainer_email='reid@reidswanson.com',
        zip_safe=False,
        packages=find_packages(),
        include_package_data=True,
        license='lgpl-v3',
        url='https://bitbucket.org/reidswanson/scrachy',
        classifiers=['Development Status :: 4 - Beta',
                     'Intended Audience :: Science/Research',
                     'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
                     'Natural Language :: English',
                     'Operating System :: MacOS',
                     'Operating System :: Microsoft :: Windows',
                     'Operating System :: POSIX',
                     'Operating System :: Unix',
                     'Programming Language :: Python :: 3.11'])



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
Some utility classes for sending messages between the
:class:`~scrachy.middleware.selenium.AsyncSeleniumMiddleware` and the
:mod:`~scrachy.cli.webdriver_server`. It also includes the primary
functionality for processing requests with Selenium. Each Selenium middleware
is a thin wrapper around these functions.
"""

from __future__ import annotations

# Python Modules
import logging

from typing import Optional, Type, cast

# 3rd Party Modules
from scrapy.http import Request, HtmlResponse
from scrapy.utils.misc import load_object
from selenium import webdriver
from selenium.common import TimeoutException
from selenium.webdriver.chromium.options import ChromiumOptions
from selenium.webdriver.common.options import ArgOptions
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support.wait import WebDriverWait

# Project Modules
from scrachy.http_ import SeleniumRequest
from scrachy.settings.defaults.selenium import WebDriverName


log = logging.getLogger(__name__)


webdriver_import_base = 'selenium.webdriver'


class BufferIncompleteError(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class ShutdownRequest:
    def __int__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class UnknownMessageType:
    def __init__(self, message_type: str, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.message_type = message_type


class UnhandledError:
    def __init__(self, exception: Exception, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.exception = exception


def initialize_driver(driver_name: WebDriverName, options: list[str], extensions: list[str]) -> WebDriver:
    driver_name: WebDriverName = driver_name
    driver_cls: Type[WebDriver] = load_object(f'{webdriver_import_base}.{driver_name}')
    driver_options: ArgOptions = load_object(f'{webdriver_import_base}.{driver_name}Options')()

    for option in options:
        driver_options.add_argument(option)

    # Chrome loads the extensions from the options
    if driver_name == 'Chrome' and extensions:
        driver_options = cast(ChromiumOptions, driver_options)
        for extension in extensions:
            driver_options.add_extension(extension)  # noqa

    driver = driver_cls(options=driver_options)

    # Firefox appears to lod the extension directly from the driver
    if driver_name == 'Firefox' and extensions:
        for extension in extensions:
            driver = cast(webdriver.Firefox, driver)
            driver.install_addon(extension, temporary=True)

    return driver


def process_request(driver: WebDriver, request: Request) -> Optional[HtmlResponse]:
    if not isinstance(request, SeleniumRequest):
        # Let some other downloader handle this request
        return None

    request = cast(SeleniumRequest, request)

    driver.get(request.url)

    set_cookies(driver, request)
    wait_for_page(driver, request)
    take_screenshot(driver, request)

    execute_script(driver, request)

    response = make_response(driver, request)

    return response


def set_cookies(driver: WebDriver, request: SeleniumRequest):
    for cookie_name, cookie_value in request.cookies.items():
        driver.add_cookie({'name': cookie_name, 'value': cookie_value})


def wait_for_page(driver: WebDriver, request: SeleniumRequest):
    if request.wait_until:
        try:
            WebDriverWait(
                driver,
                request.wait_timeout
            ).until(
                request.wait_until
            )
        except TimeoutException as e:
            log.error(f"Wait condition timed out for url: '{request.url}'")
            raise e


def take_screenshot(driver: WebDriver, request: SeleniumRequest):
    if request.screenshot:
        request.meta['screenshot'] = driver.get_screenshot_as_png()


def make_response(driver: WebDriver, request: SeleniumRequest) -> HtmlResponse:
    return HtmlResponse(
        url=driver.current_url,
        body=driver.page_source,
        encoding='utf-8',
        request=request
    )


def execute_script(driver: WebDriver, request: SeleniumRequest):
    if request.script_executor is not None:
        request.meta['script_result'] = request.script_executor(driver, request)
